from django.urls import path
from django.http import JsonResponse, HttpResponseNotAllowed
from django.shortcuts import render
from django.conf import settings
from banjo import http
from banjo import views
from banjo.forms import ApiRouteForm
import json

urlpatterns = [
    path(settings.API_PREFIX, views.api, name="api"),
    path(settings.API_PREFIX + ".json", views.api_json, name="api_json"),
]
user_defined_routes = []

class JsonRequiresDict(TypeError):
    pass

def validate_args(args):
    """Checks that args (the param type signature) is valid.
    """
    if args is None: return
    allowed_types = [str, bool, int, float]
    if not isinstance(args, dict):
        raise ValueError("args must be a dict if provided")
    for param, type_ in args.items():
        if type_ not in allowed_types:
            raise ValueError("param {} in args must be one of: {}".format(
                    param, ', '.join(str(t) for t in allowed_types)))

def create_view(fn, method, args=None):
    """Creates a django view function.
    The view function checks that the HTTP method is correct, 
    extracts the request's params, passes them to ``fn``, 
    and returns the result. Args, if provided, should be 
    a dict of ``{param_name: type}``, where type is in 
    ``[str, bool, int, float]``.
    """
    method = method.upper()
    def view(request):
        if request.method != method:
            return HttpResponseNotAllowed([method])  
        if args:
            if request.headers.get('Content-Type') == 'application/json':
                data = json.loads(request.body.decode("utf-8"))
            elif method == 'GET':
                data = request.GET
            else:
                data = request.POST
            FormClass = ApiRouteForm.for_args(args)
            form = FormClass(data)
            if form.is_valid():
                try:
                    result = fn(form.cleaned_data)
                    if not isinstance(result, dict):
                        raise JsonRequiresDict(f"Banjo views must return a dict, not {result}.")
                    return JsonResponse(result)
                except http.BadRequest as e:
                    return JsonResponse({'error': str(e)}, status=e.status_code)
            else:
                return JsonResponse({'errors': form.errors}, status=400)
        else:
            try:
                result = fn({})
                if not isinstance(result, dict):
                    raise JsonRequiresDict(f"Banjo views must return a dict, not {result}.")
                return JsonResponse(result)
            except http.BadRequest as e:
                return JsonResponse({'error': str(e)}, status=e.status_code)

    view.__name__ = fn.__name__
    view.__doc__ = fn.__doc__
    view.method = method
    validate_args(args)
    view.args = args or {}
    return view

def create_api_view(url, fn, method, args=None):
    """Creates an API view.
    An API view responds to GET requests with a HTML template--
    the response data for GET routes and a form for POST routes.
    For POST routes, the view processes the form.
    """
    method = method.upper()
    fn.method = method
    def api_view(request):
        fn.args = args or {}
        route = views.describe_route(path(url, fn, name=fn.__name__))

        if method == "GET" and request.method == "GET":
            if args:
                form = ApiRouteForm.for_args(args)(request.GET)
                if form.is_valid():
                    try:
                        result = fn(form.cleaned_data)
                    except http.BadRequest as e:
                        result = {'error': str(e), 'status_code': e.status_code}
                else:
                    result = None
            else:
                form = None
                try:
                    result = fn({})
                except http.BadRequest as e:
                    result = {'error': str(e), 'status_code': e.status_code}
            return render(request, "banjo/api_get.html", {
                "route": route,
                "result": json.dumps(result, indent=True) if result else None,
                "form": form,
            })
        if method == "POST":
            if request.method == "POST":
                if args:
                    form = ApiRouteForm.for_args(args)(request.POST)
                    if form.is_valid():
                        try:
                            result = fn(form.cleaned_data)
                        except http.BadRequest as e:
                            result = {'error': str(e), 'status_code': e.status_code}
                        form = ApiRouteForm.for_args(args)()
                    else:
                        result = None
                else:
                    form = None
                    try:
                        result = fn({})
                    except http.BadRequest as e:
                        result = {'error': str(e), 'status_code': e.status_code}
            else:
                if args:
                    form = ApiRouteForm.for_args(args)()
                else:
                    form = None
                result = None

            return render(request, "banjo/api_post.html", {
                "route": route,
                "result": json.dumps(result, indent=True) if result else None,
                "form": form,
            })

        return HttpResponseNotAllowed([request.method])

    api_view.__name__ = fn.__name__
    api_view.__doc__ = fn.__doc__
    validate_args(args)
    api_view.args = args or {}
    return api_view
    
def route_get(url, args=None):
    """A decorator which registers a HTTP GET route in the Banjo app.
    """
    def bind_url_to_view(fn):
        view = create_view(fn, "GET", args=args)
        api_view = create_api_view(url, fn, "GET", args=args)
        urlpatterns.append(path(url, view, name=view.__name__))
        urlpatterns.append(path(settings.API_PREFIX + '/' + url, api_view, 
                name="api_" + view.__name__))
        user_defined_routes.append(path(url, view, name=view.__name__))
        return view
    return bind_url_to_view

def route_post(url, args=None):
    """A decorator which registers a HTTP POST route in the Banjo app.
    """
    def bind_url_to_view(fn):
        view = create_view(fn, "POST", args=args)
        api_view = create_api_view(url, fn, "POST", args=args)
        urlpatterns.append(path(url, view, name=view.__name__))
        urlpatterns.append(path(settings.API_PREFIX + '/' + url, api_view, 
                name="api_" + view.__name__))
        user_defined_routes.append(path(url, view, name=view.__name__))
        return view
    return bind_url_to_view



from django.db import models
from django.db.models import Field, ManyToOneRel
from random import choice, sample

class QuerySet(models.QuerySet):
    """Extends the base QuerySey with random() and sample() methods
    """
    def random(self):
        ids = self.values_list('id', flat=True)
        if not ids:
            raise self.model.DoesNotExist(f"{self.model.__name__} is empty")
        return self.get(id=choice(ids))

    def sample(self, n):
        ids = self.values_list('id', flat=True)
        if len(ids) < n:
            raise self.model.DoesNotExist(f"Requested {n} but {self.model.__name__} only contains {len(ids)} objects.")
        return self.filter(id__in=sample([i for i in ids], n))

def serialize(obj):
    """Safely converts an object to a dict, 
    regardless of whether it implements banjo's `to_dict` interface.
    """
    if hasattr(obj, 'to_dict'):
        try:
            return obj.to_dict(with_related=False)
        except TypeError:
            return obj.to_dict()
    else:
        return {'id': obj.id}

class Model(models.Model):

    objects = models.Manager.from_queryset(QuerySet)()

    @classmethod
    def from_dict(cls, props):
        """Tries to create an instance of Model from props.
        """
        if 'id' in props:
            del props['id']
        return cls(**props)

    def to_dict(self, with_related=True):
        """Returns a json representation of the Model.
        """
        d = {}
        for field in self._meta.get_fields():
            if with_related and isinstance(field, models.ForeignKey):
                related_object = getattr(self, field.name)
                d[field.name] = serialize(related_object)
            elif isinstance(field, Field):
                d[field.name] = field.value_from_object(self)
            elif with_related and isinstance(field, ManyToOneRel):
                related_objects = getattr(self, field.get_accessor_name()).all()
                d[field.get_accessor_name()] = [serialize(obj) for obj in related_objects]
        return d

    def __str__(self):
        return "<{} {}>".format(self.__class__.__name__, self.to_dict())

    def __repr__(self):
        return self.__str__()

    class Meta:
        abstract = True
        app_label = "app"

class BooleanField(models.BooleanField):
    """A database column which stores a boolean.
    The default value is False.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = False
        models.BooleanField.__init__(self, *args, **kwargs)

class IntegerField(models.IntegerField):
    """A database column which stores an integer.
    The default value is 0.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = 0
        models.IntegerField.__init__(self, *args, **kwargs)

class FloatField(models.FloatField):
    """A database column which stores a float.
    The default value is 0.0.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = 0.0
        models.FloatField.__init__(self, *args, **kwargs)

class StringField(models.TextField):
    """A database column which stores a string.
    The default value is '', the empty string.
    """
    def __init__(self, *args, **kwargs):
        kwargs['default'] = ''
        models.TextField.__init__(self, *args, **kwargs)

class ForeignKey(models.ForeignKey):
    """A database column which links a model to another model.
    """
    def __init__(self, *args, **kwargs):
        kwargs['on_delete'] = models.CASCADE
        models.ForeignKey.__init__(self, *args, **kwargs)
